------------------------------------------------------------------------
-- Concrete representation of thunks
------------------------------------------------------------------------

-- The definitions in this module (except for the lemmas towards the
-- end of the file) are concrete, so they can be used to reason about
-- thunks (as illustrated by the lemmas). Note, however, that the code
-- in this module is _not_ run-time; it is only intended to be used in
-- proofs.

module Thunk.Concrete where

open import Data.Nat
open import Data.Product hiding (swap)
open import Relation.Binary.PropositionalEquality
open ≡-Reasoning
open import Data.Function

open import Thunk.Context

infix  5 _[_]
infixl 1 _>>=_
infix  0 √_

------------------------------------------------------------------------
-- Definitions

mutual

  Thunk : ℕ -> Set -> Set
  Thunk n a = Thunk' a n

  data Thunk' (a : Set) : ℕ -> Set where
    return : a -> Thunk zero a
    √_     : forall {n} -> Thunk n a -> Thunk (suc n) a

force : forall {n a} -> Thunk n a -> a
force (return x) = x
force (√ x)      = force x

map : forall {n a b} -> (a -> b) -> Thunk n a -> Thunk n b
map f (return x) = return (f x)
map f (√ x)      = √ map f x

_>>=_ :  forall {n₁ n₂ a b}
     -> Thunk n₁ a -> (a -> Thunk n₂ b) -> Thunk (n₁ + n₂) b
return x >>= f = f x
(√ x)    >>= f = √ x >>= f

return⋆ : forall {a} n -> a -> Thunk n a
return⋆ zero    x = return x
return⋆ (suc n) x = √ return⋆ n x

swap :  forall {a n₁ n₂}
     -> Thunk n₁ (Thunk n₂ a) -> Thunk n₂ (Thunk n₁ a)
swap (return x) = map return x
swap (√ x)      = map √_ (swap x)

_[_] : Ctxt -> Set -> Set
_[_] = fill Thunk

pay :  forall {n₁ a}
       (C : Ctxt) (n₂ : ℕ)
    -> C [ Thunk n₁ a ] -> Thunk n₂ (C [ Thunk (n₁ ∸ n₂) a ])
pay •             zero     x          = return x
pay •             (suc n₂) (return x) = return⋆ (suc n₂) (return x)
pay •             (suc n₂) (√ x)      = √ pay • n₂ x
pay (const• b)    n₂       x          = return⋆ n₂ x
pay (Thunk• n C)  n₂       x          = swap $ map (pay C n₂) x
pay (b ×• C)      n₂       (x , y)    = map (\y -> x , y) (pay C n₂ y)
pay (C •× b)      n₂       (x , y)    = map (\x -> x , y) (pay C n₂ x)

------------------------------------------------------------------------
-- Some properties

abstract

  prop-force-return⋆ :  forall {a} n (x : a)
                     -> force (return⋆ n x) ≡ x
  prop-force-return⋆ zero    x = ≡-refl
  prop-force-return⋆ (suc n) x = prop-force-return⋆ n x

  prop-return⋆-force :  forall {a n} (x : Thunk n a)
                     -> return⋆ n (force x) ≡ x
  prop-return⋆-force (return x) = ≡-refl
  prop-return⋆-force (√ x)      = ≡-cong √_ $ prop-return⋆-force x

  prop-map-return⋆ :  forall {a b} n (f : a -> b) x
                   -> map f (return⋆ n x) ≡ return⋆ n (f x)
  prop-map-return⋆ zero    f x = ≡-refl
  prop-map-return⋆ (suc n) f x = ≡-cong √_ $ prop-map-return⋆ n f x

  map-cong :  forall {a b n} {x₁ x₂ : Thunk n a} {f₁ f₂ : a -> b}
           -> f₁ ≗ f₂ -> x₁ ≡ x₂ -> map f₁ x₁ ≡ map f₂ x₂
  map-cong {x₁ = return x} eqf ≡-refl = ≡-cong return (eqf x)
  map-cong {x₁ = √ x}      eqf ≡-refl = ≡-cong √_ $
    map-cong {x₁ = x} eqf ≡-refl

  map-functorial
    :  forall {a b c n} (f : b -> c) (g : a -> b) (x : Thunk n a)
    -> map f (map g x) ≡ map (f ∘ g) x
  map-functorial f g (return x) = ≡-refl
  map-functorial f g (√ x)      = ≡-cong √_ $ map-functorial f g x

  prop-swap-map
    :  forall {a m} n
    -> (x : Thunk m a)
    -> swap (return⋆ n x) ≡ map (return⋆ n) x
  prop-swap-map zero    x = ≡-refl
  prop-swap-map (suc n) x =
                                 begin
    swap (return⋆ (suc n) x)
                                 ≡⟨ byDef ⟩
    map √_ (swap (return⋆ n x))
                                 ≡⟨ ≡-cong (map √_) (prop-swap-map n x) ⟩
    map √_ (map (return⋆ n) x)
                                 ≡⟨ map-functorial √_ (return⋆ n) x ⟩
    map (√_ ∘ return⋆ n) x
                                 ∎
