-- # import eq
-- # import prelude
-- # import monoid

data ℕ : Set where
  zero : ℕ
  suc  : ℕ -> ℕ

infixl 7 *
infixl 6 +
infixl 6 ∸

(+) : ℕ -> ℕ -> ℕ
zero  + n = n
suc m + n = suc (m + n)

(*) : ℕ -> ℕ -> ℕ
zero  * n = zero
suc m * n = n + m * n

double : ℕ -> ℕ
double zero    = 0
double (suc n) = suc (suc (double n))

diff : ℕ -> ℕ -> ℕ
diff zero    n       = n
diff m       zero    = m
diff (suc m) (suc n) = diff m n

pred : ℕ -> ℕ
pred zero    = zero
pred (suc n) = n

-- n ∸ m | m <= n    = n - m
--       | otherwise = 0

(∸) : ℕ -> ℕ -> ℕ
n       ∸ zero    = n
zero    ∸ (suc _) = zero
(suc m) ∸ (suc n) = m ∸ n

-- To simplify partial applications.

subtract : ℕ -> ℕ -> ℕ
subtract m n = n ∸ m

-- ---------------------------------------------------------------------
-- Lemmas

castMul (i, j, m, n : ℕ)
  : i ≡ j -> m ≡ n -> i * m ≡ j * n
castMul refl refl = refl

absurdNat (n : ℕ) (whatever : Set)
  : suc n ≡ zero -> whatever
absurdNat ()

addRightZero (n : ℕ) : n + 0 ≡ n
addRightZero {zero}  = refl
addRightZero {suc n} = cong suc addRightZero

addRightZero' : (n : ℕ) -> n + 0 == n
addRightZero' _ = toPropEq addRightZero

rewrite addRightZero'

addRightSuc (m, n : ℕ) : m + suc n ≡ suc (m + n)
addRightSuc {zero}  {n} = refl
addRightSuc {suc m} {n} = cong suc addRightSuc

addRightSuc' : (m, n : ℕ) -> m + suc n == suc (m + n)
addRightSuc' _ _ = toPropEq addRightSuc

rewrite addRightSuc'

addCommutative (m, n : ℕ) : m + n ≡ n + m
addCommutative {0}     = refl
addCommutative {suc m} = cong suc addCommutative

addAssociative'' (i, j : ℕ)
  : (f : ℕ -> ℕ) -> (\k -> f (i + (j + k))) ≡ \k -> f ((i + j) + k)
addAssociative'' {zero}  _ = refl
addAssociative'' {suc _} f = addAssociative'' (\x -> f (suc x))

addAssociative' (i, j : ℕ) : (\k -> i + (j + k)) ≡ (\k -> (i + j) + k)
addAssociative' = addAssociative'' (\x -> x)

addAssociative (k, i, j : ℕ) : i + (j + k) ≡ (i + j) + k
addAssociative {k} = cong (\f -> f k) addAssociative'

mulRightZero (n : ℕ) : n * 0 ≡ 0
mulRightZero {zero}  = refl
mulRightZero {suc _} = mulRightZero

mulRightZero' : (n : ℕ) -> n * 0 == 0
mulRightZero' _ = toPropEq mulRightZero

rewrite mulRightZero'

mulRightOne (n : ℕ) : n * 1 ≡ n
mulRightOne {zero}  = refl
mulRightOne {suc n} = cong suc mulRightOne

mulDistribLeftAdd (i, j, k : ℕ) : i * (j + k) ≡ i * j + i * k
mulDistribLeftAdd {zero}  {j} {k} = refl
mulDistribLeftAdd {suc i} {j} {k} =
                                   -- (j + k) + (i * (j + k))
  cong (\n -> j + k + n)
        mulDistribLeftAdd
    `trans`                        -- (j + k) + (i * j + i * k)
  addAssociative
    `trans`                        -- ((j + k) + i * j) + i * k
  cong (\n -> n + i * k)
        (sym addAssociative)
    `trans`                        -- (j + (k + i * j)) + i * k
  cong (\n -> j + n + i * k)
        addCommutative
    `trans`                        -- (j + (i * j + k)) + i * k
  cong (\n -> n + i * k)
        addAssociative
    `trans`                        -- ((j + i * j) + k) + i * k
  sym addAssociative
                                   -- (j + i * j) + (k + i * k)

--   suc i * (j + k)
-- =
--   (j + k) + i * (j + k)
-- = { ... }
--   (j + i * j) + (k + i * k)
-- =
--   suc i * j + suc i * k

mulCommutative (m, n : ℕ) : m * n ≡ n * m
mulCommutative {zero}  {_} = refl
mulCommutative {suc m} {n} =
  sym (mulDistribLeftAdd
         `trans`
       cong (\k -> k + (n * m)) mulRightOne
         `trans`
       cong (\k -> n + k) mulCommutative
      )

--   suc m * n
-- =
--   n + m * n
-- = { IH }
--   n + n * m
-- = { mulRightOne }
--   n * 1 + n * m
-- = { mulDistribLeftAdd }
--   n * (1 + m)

mulRightSuc (m, n : ℕ) : m * suc n ≡ m + n * m
mulRightSuc = mulCommutative

--   n * suc m
-- = { mulCommutative }
--   suc m * n
-- =
--   n + m * n

mulRightSuc' : (m, n : ℕ) -> m * suc n ≡ m + m * n
mulRightSuc' m n = mulRightSuc `trans` congBin (+) refl mulCommutative

mulDistribRightAdd (i, j, k : ℕ) : (i + j) * k ≡ i * k + j * k
mulDistribRightAdd {i} {j} {k} =
  mulCommutative
    `trans`
  mulDistribLeftAdd
    `trans`
  cong (\n -> n + k * j) mulCommutative
    `trans`
  cong (\n -> i * k + n) mulCommutative

mulAssociative (i, j, k : ℕ) : i * (j * k) ≡ (i * j) * k
mulAssociative {zero}  {j} {k} = refl
mulAssociative {suc i} {j} {k} =
  cong (\n -> j * k + n) mulAssociative
    `trans`
  sym mulDistribRightAdd

--   suc i * (j * k)
-- =
--   j * k + i * (j * k)
-- =
--   j * k + (i * j) * k
-- = { mulDistribRightAdd }
--   (j + i * j) * k
-- =
--   (suc i * j) * k

doubleMul (m : ℕ) : double m ≡ 2 * m
doubleMul {zero}  = refl
doubleMul {suc m} = cong (suc ∘ suc) doubleMul

subtractLeftZero (m : ℕ) : 0 ∸ m ≡ 0
subtractLeftZero {0}     = refl
subtractLeftZero {suc _} = refl

subtractLeftZero' : (m : ℕ) -> 0 ∸ m == 0
subtractLeftZero' _ = toPropEq subtractLeftZero

rewrite subtractLeftZero'

subtractLinear (i, j, k : ℕ) : (k ∸ j) ∸ i ≡ k ∸ (i + j)
subtractLinear {i}     {0}     {k} = refl
subtractLinear {i} {suc j}     {0} = refl
subtractLinear {i} {suc j} {suc k} = subtractLinear {i} {j} {k}

subtractPlusInverses (i, j : ℕ) : (j + i) ∸ i ≡ j
subtractPlusInverses {0}     = refl
subtractPlusInverses {suc i} = subtractPlusInverses {i}

subtractPlusInverses' (i, j : ℕ) : (i + j) ∸ i ≡ j
subtractPlusInverses' {0}     = refl
subtractPlusInverses' {suc i} = subtractPlusInverses' {i}

subtractSelf' (n, m : ℕ) : n ∸ (n + m) ≡ 0
subtractSelf' {0}     = refl
subtractSelf' {suc n} = subtractSelf' {n}

subtractSelf (n : ℕ) : n ∸ n ≡ 0
subtractSelf = subtractSelf' {_} {0}

prop_sub_swap_sub (m, n : ℕ) : (m ∸ n) ∸ (n ∸ m) ≡ m ∸ n
prop_sub_swap_sub {0}             = refl
prop_sub_swap_sub {suc m} {0}     = refl
prop_sub_swap_sub {suc m} {suc n} = prop_sub_swap_sub {m} {n}

prop_sub_swap_plus (m, n : ℕ) : (m ∸ n) + (n ∸ m) ≡ m `diff` n
prop_sub_swap_plus {0}             = refl
prop_sub_swap_plus {suc m} {0}     = refl
prop_sub_swap_plus {suc m} {suc n} = prop_sub_swap_plus {m} {n}

prop_sub_swap_sub_sub (x, m, y : ℕ)
  : x ∸ m + (y ∸ (m ∸ x)) ≡ (x + y) ∸ m
prop_sub_swap_sub_sub {0}             = refl
prop_sub_swap_sub_sub {suc x} {0}     = refl
prop_sub_swap_sub_sub {suc x} {suc m} = prop_sub_swap_sub_sub {x} {m}

------------------------------------------------------------------------
-- Monoids

natPlus : Monoid ℕ
natPlus = struct
  zero      = 0
  (+)       = (+)
  assoc     = addAssociative
  leftUnit  = refl
  rightUnit = refl

natMul : Monoid ℕ
natMul = struct
  zero      = 1
  (+)       = (*)
  assoc     = mulAssociative
  leftUnit  = refl
  rightUnit = mulRightOne
