------------------------------------------------------------------------
-- Erasure-propagating solver
------------------------------------------------------------------------

-- Automatic solver for equalities of the form
--
--   ⌜ f x y ⌝ ≡ ⌜ f ⌝ ⌜ x ⌝ ⌜ y ⌝.
--
-- The two sides are simplified by pushing the erasure functions as
-- far down the syntax trees as possible, and then the terms are
-- compared for definitional equality.

-- For example usage, see OperationalSemantics.Sound.

module Erasure.Solver where

open import Kind
open import TypeSystem
open import Erasure
open import Value
open import OperationalSemantics.Binding
open Heap↦
open Heap↦Erasure
open import Substitution
open VarSubst
open import Erasure.Lemmas
open import Relation.Binary.PropositionalEquality
open ≡-Reasoning
open import Data.Nat

infixl 70 _|·|_ _|∘ˢ|_
infix  70 _|↑|
infixl 48 _|/⊢|_ _|/⊢⇓|_ _|/∋|_ _|/↦|_
infix  30 _↓ _↓⇒ _↓⊢ _↓⇓ |↦|_ |↦ⁿ|_
infixl 20 _|▷|_ _|▷ʰ|_
infix  10 _|⇒|_ _|⊢|_ _|⊢⇓|_ _|∋|_ _|↦|_

------------------------------------------------------------------------
-- Representations

-- Representation of types for representable terms.

data |Ty| (k : Kind) : Set where
  _|⇒|_  : Ctxt k -> Ctxt k -> |Ty| k
  _|⊢|_  : Ctxt k -> Ty k   -> |Ty| k
  _|⊢⇓|_ : Ctxt k -> Ty k   -> |Ty| k
  _|∋|_  : Ctxt k -> Ty k   -> |Ty| k
  _|↦|_  : Ctxt k -> Ty k   -> |Ty| k
  |Heap| : Ctxt k           -> |Ty| k

-- Interpretation of the types.

⟦_⟧⋆ : forall {k} -> |Ty| k -> Set
⟦ Γ |⇒|  Δ ⟧⋆ = Γ ⇒  Δ
⟦ Γ |⊢|  σ ⟧⋆ = Γ ⊢  σ
⟦ Γ |⊢⇓| σ ⟧⋆ = Γ ⊢⇓ σ
⟦ Γ |∋|  σ ⟧⋆ = Γ ∋  σ
⟦ Γ |↦|  σ ⟧⋆ = Γ ↦  σ
⟦ |Heap| Γ ⟧⋆ = Heap Γ

-- Representation of terms between which certain equalitites can be
-- proved using prove.

data |Tm| : forall {k} -> |Ty| k -> Set where

  -- Anything with a representable type can be lifted.

  lift            :  forall {k} {σ : |Ty| k} -> ⟦ σ ⟧⋆ -> |Tm| σ

  -- Substitutions.

  _|▷|_           : forall {k} {Γ Δ : Ctxt k} {τ}
                  -> |Tm| (Γ |⇒| Δ) -> |Tm| (Δ |∋| τ)
                  -> |Tm| (Γ ▻ τ |⇒| Δ)
  |wk|            :  forall {k} {Γ : Ctxt k} {σ}
                  -> |Tm| (Γ |⇒| Γ ▻ σ)
  |wk⇒|           :  forall {k} {Γ Δ : Ctxt k} {σ}
                  -> |Tm| (Γ |⇒| Δ) -> |Tm| (Γ |⇒| Δ ▻ σ)
  |sub|           :  forall {k} {Γ : Ctxt k} {σ}
                  -> |Tm| (Γ |∋| σ) -> |Tm| (Γ ▻ σ |⇒| Γ)
  _|↑|            :  forall {k} {Γ Δ : Ctxt k} {σ}
                  -> |Tm| (Γ |⇒| Δ) -> |Tm| (Γ ▻ σ |⇒| Δ ▻ σ)
  |idˢ|           :  forall {k} {Γ : Ctxt k}
                  -> |Tm| (Γ |⇒| Γ)
  _|∘ˢ|_          :  forall {k} {Γ Δ X : Ctxt k}
                  -> |Tm| (Γ |⇒| Δ) -> |Tm| (Δ |⇒| X) -> |Tm| (Γ |⇒| X)
  |⌜_⌝⇒|          :  forall {Γ Δ : Ctxt thunked}
                  -> |Tm| (Γ |⇒| Δ) -> |Tm| (⌜ Γ ⌝▻ |⇒| ⌜ Δ ⌝▻)

  -- Terms.

  _|/⊢|_          :  forall {k} {Γ Δ : Ctxt k} {σ}
                  -> |Tm| (Γ |⊢| σ) -> |Tm| (Γ |⇒| Δ) -> |Tm| (Δ |⊢| σ)
  |var|           :  forall {k} {Γ : Ctxt k} {σ}
                  -> |Tm| (Γ |∋| σ) -> |Tm| (Γ |⊢| σ)
  _|·|_           :  forall {k} {Γ : Ctxt k} {σ τ}
                  -> |Tm| (Γ |⊢| σ → τ) -> |Tm| (Γ |⊢| σ) -> |Tm| (Γ |⊢| τ)
  |natrec|        :  forall {k} {Γ : Ctxt k} {n σ}
                  -> |Tm| (Γ |⊢| Thunk? n σ)
                  -> |Tm| (Γ ▻ Nat ▻ Thunk? (suc' k n) σ |⊢| Thunk? n σ)
                  -> |Tm| (Γ |⊢| Nat → Thunk? (suc' k n) σ)
  |valToTm|       :  forall {k} {Γ : Ctxt k} {σ}
                  -> |Tm| (Γ |⊢⇓| σ) -> |Tm| (Γ |⊢| σ)
  |≡-subst⊢|      :  forall {k} {Γ : Ctxt k} {σ₁ σ₂}
                  -> σ₁ ≡ σ₂ -> |Tm| (Γ |⊢| σ₁) -> |Tm| (Γ |⊢| σ₂)
  |gpay|          :  forall {Γ σ n₁} C n₂
                  -> |Tm| (Γ |⊢| C [ Thunk n₁ σ ])
                  -> |Tm| (Γ |⊢| Thunk n₂ (C [ Thunk (n₁ ∸ n₂) σ ]))
  |⌜_⌝|           :  forall {Γ σ}
                  -> |Tm| (Γ |⊢| σ) -> |Tm| (⌜ Γ ⌝▻ |⊢| ⌜ σ ⌝⋆)

  -- Values.

  |natrec⇓|       :  forall {Γ : Ctxt simple} {σ}
                  -> |Tm| (Γ |∋| σ) -> |Tm| (Γ ▻ Nat ▻ σ |⊢| σ)
                  -> |Tm| (Γ |⊢⇓| Nat → σ)
  |drop-return⋆⇓| :  forall {Γ τ} ns
                  -> |Tm| (Γ |⊢⇓| Thunks ns τ) -> |Tm| (Γ |⊢⇓| τ)
  |gwait⇓|        :  forall {Γ σ n₁} C n₂
                  -> |Tm| (Γ |⊢⇓| C [ Thunk n₁ σ ])
                  -> |Tm| (Γ |⊢⇓| C [ Thunk (n₂ + n₁) σ ])
  |≡-subst⇓|      :  forall {Γ : Ctxt simple} {σ₁ σ₂}
                  -> σ₁ ≡ σ₂ -> |Tm| (Γ |⊢⇓| σ₁) -> |Tm| (Γ |⊢⇓| σ₂)
  _|/⊢⇓|_         :  forall {k} {Γ Δ : Ctxt k} {σ}
                  -> |Tm| (Γ |⊢⇓| σ) -> |Tm| (Γ |⇒| Δ) -> |Tm| (Δ |⊢⇓| σ)
  |⌜_⌝⇓|          :  forall {Γ σ}
                  -> |Tm| (Γ |⊢⇓| σ) -> |Tm| (⌜ Γ ⌝▻ |⊢⇓| ⌜ σ ⌝⋆)

  -- Variables.

  |gwait∋|        :  forall {Γ σ n₁} C n₂
                  -> |Tm| (Γ |∋| C [ Thunk n₁ σ ])
                  -> |Tm| (Γ |∋| C [ Thunk (n₂ + n₁) σ ])
  |≡-subst∋|      :  forall {k} {Γ : Ctxt k} {σ₁ σ₂}
                  -> σ₁ ≡ σ₂ -> |Tm| (Γ |∋| σ₁) -> |Tm| (Γ |∋| σ₂)
  _|/∋|_          :  forall {k} {Γ Δ : Ctxt k} {σ}
                  -> |Tm| (Γ |∋| σ) -> |Tm| (Γ |⇒| Δ) -> |Tm| (Δ |∋| σ)
  |⌜_⌝∋|          :  forall {Γ σ}
                  -> |Tm| (Γ |∋| σ) -> |Tm| (⌜ Γ ⌝▻ |∋| ⌜ σ ⌝⋆)

  -- Heaps.

  _|▷ʰ|_          :  forall {k} {Γ : Ctxt k} {σ}
                  -> |Tm| (|Heap| Γ) -> |Tm| (Γ |↦| σ)
                  -> |Tm| (|Heap| (Γ ▻ σ))
  |⌜_⌝ʰ|          :  forall {Γ}
                  -> |Tm| (|Heap| Γ) -> |Tm| (|Heap| ⌜ Γ ⌝▻)

  -- Bindings.

  |↦|_            :  forall {k} {Γ : Ctxt k} {σ}
                  -> |Tm| (Γ |⊢| σ) -> |Tm| (Γ |↦| σ)
  |↦ⁿ|_           :  forall {Γ n σ}
                  -> |Tm| (Γ |⊢| Thunk n σ) -> |Tm| (Γ |↦| σ)
  _|/↦|_          :  forall {k} {Γ Δ : Ctxt k} {σ}
                  -> |Tm| (Γ |↦| σ) -> |Tm| (Γ |⇒| Δ) -> |Tm| (Δ |↦| σ)
  |⌜_⌝↦|          :  forall {Γ σ}
                  -> |Tm| (Γ |↦| σ) -> |Tm| (⌜ Γ ⌝▻ |↦| ⌜ σ ⌝⋆)

-- Some abbreviations.

lift⇒ : forall {k} {Γ Δ : Ctxt k} -> ⟦ Γ |⇒| Δ ⟧⋆ -> |Tm| (Γ |⇒| Δ)
lift⇒ = lift {σ = _ |⇒| _}

liftʰ : forall {k} {Γ : Ctxt k} -> ⟦ |Heap| Γ ⟧⋆ -> |Tm| (|Heap| Γ)
liftʰ = lift {σ = |Heap| _}

-- Interpretation of the terms.

⟦_⟧ : forall {k σ} -> |Tm| {k} σ -> ⟦ σ ⟧⋆
⟦ lift x ⟧               = x
⟦ ρ |▷| x ⟧              = ⟦ ρ ⟧ ▷ ⟦ x ⟧
⟦ |wk| ⟧                 = wk
⟦ |wk⇒| ρ ⟧              = wk⇒ ⟦ ρ ⟧
⟦ |sub| x ⟧              = sub ⟦ x ⟧
⟦ ρ |↑| ⟧                = ⟦ ρ ⟧ ↑
⟦ |idˢ| ⟧                = idˢ
⟦ ρ₁ |∘ˢ| ρ₂ ⟧           = ⟦ ρ₁ ⟧ ∘ˢ ⟦ ρ₂ ⟧
⟦ |⌜ ρ ⌝⇒| ⟧             = ⌜ ⟦ ρ ⟧ ⌝⇒
⟦ e |/⊢| ρ ⟧             = ⟦ e ⟧ /⊢ ⟦ ρ ⟧
⟦ |var| x ⟧              = var ⟦ x ⟧
⟦ t₁ |·| t₂ ⟧            = ⟦ t₁ ⟧ · ⟦ t₂ ⟧
⟦ |natrec| t₁ t₂ ⟧       = natrec ⟦ t₁ ⟧ ⟦ t₂ ⟧
⟦ |valToTm| v ⟧          = valToTm ⟦ v ⟧
⟦ |≡-subst⊢| eq t ⟧      = ≡-subst (_⊢_ _) eq ⟦ t ⟧
⟦ |gpay| C n₂ t ⟧        = gpay C n₂ ⟦ t ⟧
⟦ |⌜ e ⌝| ⟧              = ⌜ ⟦ e ⟧ ⌝
⟦ |natrec⇓| x₁ v₂ ⟧      = natrec⇓ ⟦ x₁ ⟧ ⟦ v₂ ⟧
⟦ |drop-return⋆⇓| ns v ⟧ = drop-return⋆⇓ ns ⟦ v ⟧
⟦ |gwait⇓| C m v ⟧       = gwait⇓ C m ⟦ v ⟧
⟦ |≡-subst⇓| eq v ⟧      = ≡-subst (_⊢⇓_ _) eq ⟦ v ⟧
⟦ v |/⊢⇓| ρ ⟧            = ⟦ v ⟧ /⊢⇓ ⟦ ρ ⟧
⟦ |⌜ v ⌝⇓| ⟧             = ⌜ ⟦ v ⟧ ⌝⇓
⟦ |gwait∋| C m x ⟧       = gwait∋ C m ⟦ x ⟧
⟦ |≡-subst∋| eq x ⟧      = ≡-subst (_∋_ _) eq ⟦ x ⟧
⟦ x |/∋| ρ ⟧             = ⟦ x ⟧ /∋ ⟦ ρ ⟧
⟦ |⌜ x ⌝∋| ⟧             = ⌜ ⟦ x ⟧ ⌝∋
⟦ Σ |▷ʰ| b ⟧             = ⟦ Σ ⟧ ▷ʰ ⟦ b ⟧
⟦ |⌜ Σ ⌝ʰ| ⟧             = ⌜ ⟦ Σ ⟧ ⌝ʰ
⟦ |↦|  e ⟧               = ↦  ⟦ e ⟧
⟦ |↦ⁿ| e ⟧               = ↦ⁿ ⟦ e ⟧
⟦ b |/↦| ρ ⟧             = ⟦ b ⟧ /↦ ⟦ ρ ⟧
⟦ |⌜ b ⌝↦| ⟧             = ⌜ ⟦ b ⟧ ⌝↦

------------------------------------------------------------------------
-- Simplification

mutual

  -- Simplification of terms.

  _↓ : forall {k σ} -> |Tm| {k} σ -> |Tm| σ
  (lift x)               ↓ = lift x
  (ρ |▷| x)              ↓ = (ρ ↓) |▷| (x ↓)
  |wk|                   ↓ = |wk|
  (|wk⇒| ρ)              ↓ = |wk⇒| (ρ ↓)
  (|sub| x)              ↓ = |sub| (x ↓)
  (ρ |↑|)                ↓ = (ρ ↓) |↑|
  |idˢ|                  ↓ = |idˢ|
  (ρ₁ |∘ˢ| ρ₂)           ↓ = (ρ₁ ↓) |∘ˢ| (ρ₂ ↓)
  |⌜ ρ ⌝⇒|               ↓ = ρ ↓⇒
  (t |/⊢| ρ)             ↓ = (t ↓) |/⊢| (ρ ↓)
  (|var| x)              ↓ = |var| (x ↓)
  (t₁ |·| t₂)            ↓ = (t₁ ↓) |·| (t₂ ↓)
  (|natrec| t₁ t₂)       ↓ = |natrec| (t₁ ↓) (t₂ ↓)
  (|valToTm| v)          ↓ = |valToTm| (v ↓)
  (|≡-subst⊢| eq t)      ↓ = |≡-subst⊢| eq (t ↓)
  (|gpay| C n₂ t)        ↓ = |gpay| C n₂ (t ↓)
  |⌜ e ⌝|                ↓ = e ↓⊢
  (|natrec⇓| x₁ v₂)      ↓ = |natrec⇓| (x₁ ↓) (v₂ ↓)
  (|drop-return⋆⇓| ns v) ↓ = |drop-return⋆⇓| ns (v ↓)
  (|gwait⇓| C m v)       ↓ = |gwait⇓| C m (v ↓)
  (|≡-subst⇓| eq v)      ↓ = |≡-subst⇓| eq (v ↓)
  (v |/⊢⇓| ρ)            ↓ = (v ↓) |/⊢⇓| (ρ ↓)
  |⌜ v ⌝⇓|               ↓ = v ↓⇓
  (|gwait∋| C m x)       ↓ = |gwait∋| C m (x ↓)
  (|≡-subst∋| eq x)      ↓ = |≡-subst∋| eq (x ↓)
  (x |/∋| ρ)             ↓ = (x ↓) |/∋| (ρ ↓)
  |⌜ x ⌝∋|               ↓ = x ↓∋
  (Σ |▷ʰ| b)             ↓ = (Σ ↓) |▷ʰ| (b ↓)
  |⌜ Σ ⌝ʰ|               ↓ = Σ ↓ʰ
  (|↦|  e)               ↓ = |↦|  e ↓
  (|↦ⁿ| e)               ↓ = |↦ⁿ| e ↓
  (b |/↦| ρ)             ↓ = (b ↓) |/↦| (ρ ↓)
  |⌜ b ⌝↦|               ↓ = b ↓↦

  _↓⇒ :  forall {Γ Δ : Ctxt thunked}
      -> |Tm| (Γ |⇒| Δ) -> |Tm| (⌜ Γ ⌝▻ |⇒| ⌜ Δ ⌝▻)
  (lift ρ)           ↓⇒ = |⌜ lift⇒ ρ ⌝⇒|
  (ρ |▷| x)          ↓⇒ = (ρ ↓⇒) |▷| (x ↓∋)
  |wk|               ↓⇒ = |wk|
  (|wk⇒| ρ)          ↓⇒ = |wk⇒| (ρ ↓⇒)
  (|sub| x)          ↓⇒ = |sub| (x ↓∋)
  (ρ |↑|)            ↓⇒ = (ρ ↓⇒) |↑|
  |idˢ|              ↓⇒ = |idˢ|
  (ρ₁ |∘ˢ| ρ₂)       ↓⇒ = (ρ₁ ↓⇒) |∘ˢ| (ρ₂ ↓⇒)

  _↓⊢ :  forall {Γ : Ctxt thunked} {σ}
      -> |Tm| (Γ |⊢| σ) -> |Tm| (⌜ Γ ⌝▻ |⊢| ⌜ σ ⌝⋆)
  (lift e)          ↓⊢ = |⌜ lift e ⌝|
  (e |/⊢| ρ)        ↓⊢ = (e ↓⊢) |/⊢| (ρ ↓⇒)
  (|var| x)         ↓⊢ = |var| (x ↓∋)
  (e₁ |·| e₂)       ↓⊢ = (e₁ ↓⊢) |·| (e₂ ↓⊢)
  (|natrec| t₁ t₂)  ↓⊢ = |natrec| (t₁ ↓⊢) (t₂ ↓⊢)
  (|valToTm| v)     ↓⊢ = |valToTm| (v ↓⇓)
  (|≡-subst⊢| eq t) ↓⊢ = |≡-subst⊢| (≡-cong ⌜_⌝⋆ eq) (t ↓⊢)
  (|gpay| C n₂ t)   ↓⊢ = |≡-subst⊢| (erase-lemma C) (t ↓⊢)

  _↓⇓ :  forall {Γ : Ctxt thunked} {σ}
       -> |Tm| (Γ |⊢⇓| σ) -> |Tm| (⌜ Γ ⌝▻ |⊢⇓| ⌜ σ ⌝⋆)
  (lift v)               ↓⇓ = |⌜ lift v ⌝⇓|
  (|drop-return⋆⇓| ns v) ↓⇓ = |≡-subst⇓| (Thunks-lemma ns) (v ↓⇓)
  (|gwait⇓| C m v)       ↓⇓ = |≡-subst⇓| (erase-lemma C) (v ↓⇓)
  (v |/⊢⇓| ρ)            ↓⇓ = (v ↓⇓) |/⊢⇓| (ρ ↓⇒)

  _↓∋ :  forall {Γ : Ctxt thunked} {σ}
      -> |Tm| (Γ |∋| σ) -> |Tm| (⌜ Γ ⌝▻ |∋| ⌜ σ ⌝⋆)
  (lift x)          ↓∋ = |⌜ lift x ⌝∋|
  (|gwait∋| C m x)  ↓∋ = |≡-subst∋| (erase-lemma C) (x ↓∋)
  (|≡-subst∋| eq x) ↓∋ = |≡-subst∋| (≡-cong ⌜_⌝⋆ eq) (x ↓∋)
  (x |/∋| ρ)        ↓∋ = (x ↓∋) |/∋| (ρ ↓⇒)

  _↓ʰ :  forall {Γ : Ctxt thunked}
      -> |Tm| (|Heap| Γ) -> |Tm| (|Heap| ⌜ Γ ⌝▻)
  (lift Σ)   ↓ʰ = |⌜ liftʰ Σ ⌝ʰ|
  (Σ |▷ʰ| b) ↓ʰ = (Σ ↓ʰ) |▷ʰ| (b ↓↦)

  _↓↦ :  forall {Γ : Ctxt thunked} {σ}
      -> |Tm| (Γ |↦| σ) -> |Tm| (⌜ Γ ⌝▻ |↦| ⌜ σ ⌝⋆)
  (lift b)   ↓↦ = |⌜ lift b ⌝↦|
  (|↦|  e)   ↓↦ = |↦| (e ↓⊢)
  (|↦ⁿ| e)   ↓↦ = |↦| (e ↓⊢)
  (b |/↦| ρ) ↓↦ = (b ↓↦) |/↦| (ρ ↓⇒)

abstract
 private
  mutual

    -- Simplification preserves the semantics.

    lemma : forall {k σ} (t : |Tm| {k} σ) -> ⟦ t ↓ ⟧ ≡ ⟦ t ⟧
    lemma (lift x)               = ≡-refl
    lemma (ρ |▷| x)              = ≡-cong₂ _▷_ (lemma ρ) (lemma x)
    lemma |wk|                   = ≡-refl
    lemma (|wk⇒| ρ)              = ≡-cong wk⇒ (lemma ρ)
    lemma (|sub| x)              = ≡-cong sub (lemma x)
    lemma (ρ |↑|)                = ≡-cong _↑ (lemma ρ)
    lemma |idˢ|                  = ≡-refl
    lemma (ρ₁ |∘ˢ| ρ₂)           = ≡-cong₂ _∘ˢ_ (lemma ρ₁) (lemma ρ₂)
    lemma |⌜ ρ ⌝⇒|               = lemma⇒ ρ
    lemma (e |/⊢| ρ)             = ≡-cong₂ _/⊢_ (lemma e) (lemma ρ)
    lemma (|var| x)              = ≡-cong var (lemma x)
    lemma (t₁ |·| t₂)            = ≡-cong₂ _·_ (lemma t₁) (lemma t₂)
    lemma (|natrec| t₁ t₂)       = ≡-cong₂ natrec (lemma t₁) (lemma t₂)
    lemma (|valToTm| v)          = ≡-cong valToTm (lemma v)
    lemma (|≡-subst⊢| ≡-refl t)  = lemma t
    lemma (|gpay| C n₂ t)        = ≡-cong (gpay C n₂) (lemma t)
    lemma |⌜ e ⌝|                = lemma⊢ e
    lemma (|natrec⇓| x₁ v₂)      = ≡-cong₂ natrec⇓ (lemma x₁) (lemma v₂)
    lemma (|drop-return⋆⇓| ns v) = ≡-cong (drop-return⋆⇓ ns) (lemma v)
    lemma (|gwait⇓| C m v)       = ≡-cong (gwait⇓ C m) (lemma v)
    lemma (|≡-subst⇓| ≡-refl v)  = lemma v
    lemma (v |/⊢⇓| ρ)            = ≡-cong₂ _/⊢⇓_ (lemma v) (lemma ρ)
    lemma |⌜ v ⌝⇓|               = lemma⇓ v
    lemma (|gwait∋| C m x)       = ≡-cong (gwait∋ C m) (lemma x)
    lemma (|≡-subst∋| ≡-refl x)  = lemma x
    lemma (x |/∋| ρ)             = ≡-cong₂ _/∋_ (lemma x) (lemma ρ)
    lemma |⌜ x ⌝∋|               = lemma∋ x
    lemma (Σ |▷ʰ| t)             = ≡-cong₂ _▷ʰ_  (lemma Σ) (lemma t)
    lemma |⌜ Σ ⌝ʰ|               = lemmaʰ Σ
    lemma (|↦|  e)               = ≡-cong ↦_  (lemma e)
    lemma (|↦ⁿ| e)               = ≡-cong ↦ⁿ_ (lemma e)
    lemma (b |/↦| ρ)             = ≡-cong₂ _/↦_ (lemma b) (lemma ρ)
    lemma |⌜ b ⌝↦|               = lemma↦ b

    lemma⇒ : forall {Γ Δ : Ctxt thunked} (ρ : |Tm| (Γ |⇒| Δ))
           -> ⟦ |⌜ ρ ⌝⇒| ↓ ⟧ ≡ ⟦ |⌜ ρ ⌝⇒| ⟧
    lemma⇒ (lift ρ)                = ≡-refl
    lemma⇒ (ρ |▷| x)               = ≡-cong₂ _▷_ (lemma⇒ ρ) (lemma∋ x)
    lemma⇒ (|wk|  {Γ = Γ} {σ = σ}) = ⌜wk⌝-lemma Γ σ
    lemma⇒ (|wk⇒| {σ = σ} ρ)       = begin
      ⟦ |⌜ |wk⇒| {σ = σ} ρ ⌝⇒| ↓ ⟧
                                    ≡⟨ ≡-cong wk⇒ (lemma⇒ ρ) ⟩
      wk⇒ ⟦ |⌜ ρ ⌝⇒| ⟧
                                    ≡⟨ ⌜wk⇒⌝-lemma σ ⟦ ρ ⟧ ⟩
      ⟦ |⌜ |wk⇒| {σ = σ} ρ ⌝⇒| ⟧
                                    ∎
    lemma⇒ (|sub| x) = begin
      ⟦ |⌜ |sub| x ⌝⇒| ↓ ⟧
                            ≡⟨ ≡-cong sub (lemma∋ x) ⟩
      sub ⟦ |⌜ x ⌝∋| ⟧
                            ≡⟨ ⌜sub⌝-lemma ⟦ x ⟧ ⟩
      ⟦ |⌜ |sub| x ⌝⇒| ⟧
                            ∎
    lemma⇒ (_|↑| {σ = σ} ρ) = begin
      ⟦ |⌜ _|↑| {σ = σ} ρ ⌝⇒| ↓ ⟧
                                   ≡⟨ ≡-cong _↑ (lemma⇒ ρ) ⟩
      ⟦ |⌜ ρ ⌝⇒| ⟧ ↑
                                   ≡⟨ ⌜↑⌝-lemma σ ⟦ ρ ⟧ ⟩
      ⟦ |⌜ _|↑| {σ = σ} ρ ⌝⇒| ⟧
                                   ∎
    lemma⇒ (|idˢ| {Γ = Γ}) = ⌜id⌝-lemma Γ
    lemma⇒ (ρ₁ |∘ˢ| ρ₂)    = begin
      ⟦ |⌜ ρ₁ |∘ˢ| ρ₂ ⌝⇒| ↓ ⟧
                                      ≡⟨ ≡-cong₂ _∘ˢ_ (lemma⇒ ρ₁) (lemma⇒ ρ₂) ⟩
      ⟦ |⌜ ρ₁ ⌝⇒| ⟧ ∘ˢ ⟦ |⌜ ρ₂ ⌝⇒| ⟧
                                      ≡⟨ ⌜∘⌝-lemma ⟦ ρ₁ ⟧ ⟦ ρ₂ ⟧ ⟩
      ⟦ |⌜ ρ₁ |∘ˢ| ρ₂ ⌝⇒| ⟧
                                      ∎

    lemma⊢ :  forall {Γ : Ctxt thunked} {σ} (t : |Tm| (Γ |⊢| σ))
           -> ⟦ |⌜ t ⌝| ↓ ⟧ ≡ ⟦ |⌜ t ⌝| ⟧
    lemma⊢ (lift e)   = ≡-refl
    lemma⊢ (e |/⊢| ρ) = begin
      ⟦ |⌜ e |/⊢| ρ ⌝| ↓ ⟧
                                   ≡⟨ ≡-cong₂ _/⊢_ (lemma⊢ e) (lemma⇒ ρ) ⟩
      ⟦ |⌜ e ⌝| ⟧ /⊢ ⟦ |⌜ ρ ⌝⇒| ⟧
                                   ≡⟨ ⌜/⊢⌝-lemma ⟦ e ⟧ ⟦ ρ ⟧ ⟩
      ⟦ |⌜ e |/⊢| ρ ⌝| ⟧
                                   ∎
    lemma⊢ (|var| x)        = ≡-cong var (lemma∋ x)
    lemma⊢ (e₁ |·| e₂)      = ≡-cong₂ _·_ (lemma⊢ e₁) (lemma⊢ e₂)
    lemma⊢ (|natrec| e₁ e₂) = ≡-cong₂ natrec (lemma⊢ e₁) (lemma⊢ e₂)
    lemma⊢ (|valToTm| v)    = begin
      valToTm ⟦ v ↓⇓ ⟧
                          ≡⟨ ≡-cong valToTm (lemma⇓ v) ⟩
      valToTm ⌜ ⟦ v ⟧ ⌝⇓
                          ≡⟨ ⌜valToTm⌝-lemma ⟦ v ⟧ ⟩
      ⌜ valToTm ⟦ v ⟧ ⌝
                          ∎
    lemma⊢ (|≡-subst⊢| ≡-refl t) = lemma⊢ t
    lemma⊢ (|gpay| C n₂ t)       =
      ≡-cong (≡-subst (_⊢_ _) (erase-lemma C)) (lemma⊢ t)

    lemma⇓ :  forall {Γ : Ctxt thunked} {σ} (v : |Tm| (Γ |⊢⇓| σ))
           -> ⟦ |⌜ v ⌝⇓| ↓ ⟧ ≡ ⟦ |⌜ v ⌝⇓| ⟧
    lemma⇓ (lift v)                       = ≡-refl
    lemma⇓ (|drop-return⋆⇓| {Γ = Γ} ns v) = begin
      cast ⟦ |⌜ v ⌝⇓| ↓ ⟧
                                   ≡⟨ ≡-cong cast (lemma⇓ v) ⟩
      cast ⌜ ⟦ v ⟧ ⌝⇓
                                   ≡⟨ ⌜drop-return⋆⇓⌝-lemma ns ⟦ v ⟧ ⟩
      ⌜ drop-return⋆⇓ ns ⟦ v ⟧ ⌝⇓
                                   ∎
      where cast = ≡-subst (_⊢⇓_ _) (Thunks-lemma ns)
    lemma⇓ (|gwait⇓| C m v) = begin
      cast ⟦ |⌜ v ⌝⇓| ↓ ⟧
                             ≡⟨ ≡-cong cast (lemma⇓ v) ⟩
      cast ⌜ ⟦ v ⟧ ⌝⇓
                             ≡⟨ ⌜gwait⇓⌝-lemma C m ⟦ v ⟧ ⟩
      ⌜ gwait⇓ C m ⟦ v ⟧ ⌝⇓
                             ∎
      where cast = ≡-subst (_⊢⇓_ _) (erase-lemma C)
    lemma⇓ (v |/⊢⇓| ρ) = begin
      ⟦ |⌜ v |/⊢⇓| ρ ⌝⇓| ↓ ⟧
                                     ≡⟨ ≡-cong₂ _/⊢⇓_ (lemma⇓ v) (lemma⇒ ρ) ⟩
      ⟦ |⌜ v ⌝⇓| ⟧ /⊢⇓ ⟦ |⌜ ρ ⌝⇒| ⟧
                                     ≡⟨ ⌜/⊢⇓⌝-lemma ⟦ v ⟧ ⟦ ρ ⟧ ⟩
      ⟦ |⌜ v |/⊢⇓| ρ ⌝⇓| ⟧
                                     ∎

    lemma∋ :  forall {Γ : Ctxt thunked} {σ} (x : |Tm| (Γ |∋| σ))
           -> ⟦ |⌜ x ⌝∋| ↓ ⟧ ≡ ⟦ |⌜ x ⌝∋| ⟧
    lemma∋ (lift x)         = ≡-refl
    lemma∋ (|gwait∋| C m x) =
      ≡-cong (≡-subst (_∋_ _) (erase-lemma C)) (lemma∋ x)
    lemma∋ (|≡-subst∋| ≡-refl x) = lemma∋ x
    lemma∋ (x |/∋| ρ)            = begin
      ⟦ |⌜ x |/∋| ρ ⌝∋| ↓ ⟧
                                    ≡⟨ ≡-cong₂ _/∋_ (lemma∋ x) (lemma⇒ ρ) ⟩
      ⟦ |⌜ x ⌝∋| ⟧ /∋ ⟦ |⌜ ρ ⌝⇒| ⟧
                                    ≡⟨ ⌜/∋⌝-lemma ⟦ x ⟧ ⟦ ρ ⟧ ⟩
      ⟦ |⌜ x |/∋| ρ ⌝∋| ⟧
                                    ∎

    lemmaʰ :  forall {Γ : Ctxt thunked} (Σ : |Tm| (|Heap| Γ))
           -> ⟦ |⌜ Σ ⌝ʰ| ↓ ⟧ ≡ ⟦ |⌜ Σ ⌝ʰ| ⟧
    lemmaʰ (lift Σ)   = ≡-refl
    lemmaʰ (Σ |▷ʰ| b) = ≡-cong₂ _▷ʰ_ (lemmaʰ Σ) (lemma↦ b)

    lemma↦ :  forall {Γ : Ctxt thunked} {σ} (b : |Tm| (Γ |↦| σ))
           -> ⟦ |⌜ b ⌝↦| ↓ ⟧ ≡ ⟦ |⌜ b ⌝↦| ⟧
    lemma↦ (lift b)   = ≡-refl
    lemma↦ (|↦|  e)   = ≡-cong ↦_ (lemma⊢ e)
    lemma↦ (|↦ⁿ| e)   = ≡-cong ↦_ (lemma⊢ e)
    lemma↦ (b |/↦| ρ) = begin
      ⟦ |⌜ b |/↦| ρ ⌝↦| ↓ ⟧
                                    ≡⟨ ≡-cong₂ _/↦_ (lemma↦ b) (lemma⇒ ρ) ⟩
      ⟦ |⌜ b ⌝↦| ⟧ /↦ ⟦ |⌜ ρ ⌝⇒| ⟧
                                    ≡⟨ ⌜/↦⌝-lemma ⟦ b ⟧ ⟦ ρ ⟧ ⟩
      ⟦ |⌜ b |/↦| ρ ⌝↦| ⟧
                                    ∎

------------------------------------------------------------------------
-- Prover

abstract

  -- If two terms have equal simplified syntax, then they are
  -- semantically equal.

  prove :  forall {k σ} (t₁ t₂ : |Tm| {k} σ)
        -> t₁ ↓ ≡ t₂ ↓ -> ⟦ t₁ ⟧ ≡ ⟦ t₂ ⟧
  prove t₁ t₂ eq = begin
    ⟦ t₁ ⟧
              ≡⟨ ≡-sym (lemma t₁) ⟩
    ⟦ t₁ ↓ ⟧
              ≡⟨ ≡-cong ⟦_⟧ eq ⟩
    ⟦ t₂ ↓ ⟧
              ≡⟨ lemma t₂ ⟩
    ⟦ t₂ ⟧
              ∎
