------------------------------------------------------------------------
-- Types C [ Thunk n σ ] represented as triples ⟨ C , n , σ ⟩
------------------------------------------------------------------------

-- The contribution of this module is a data type enumerating all ways
-- in which C₁ [ Thunk n₁ σ₁ ] can be equal to C₂ [ Thunk n₂ σ₂ ].

module TypeTriple where

open import Kind
open import TypeSystem

open import Data.Nat using (ℕ)
open import Relation.Binary.PropositionalEquality
  using (_≡_; ≡-refl; ≡-cong)

------------------------------------------------------------------------
-- Type triples

data TypeTriple : Set where
  ⟨_,_,_⟩ : PayCtxt -> ℕ -> Ty thunked -> TypeTriple

-- The semantics of a type triple.

⟦_⟧ : TypeTriple -> Ty thunked
⟦ ⟨ C , n , σ ⟩ ⟧ = C [ Thunk n σ ]

-- Equality between type triples.

infix 10 _≜_

data _≜_ : TypeTriple -> TypeTriple -> Set where
  ≜c-   :  forall {C n₁ σ₁ n₂ σ₂}
        -> ⟨ const● (C [ Thunk n₂ σ₂ ]) , n₁ , σ₁ ⟩ ≜ ⟨ C , n₂ , σ₂ ⟩
  ≜-c   :  forall {C n₁ σ₁ n₂ σ₂}
        -> ⟨ C , n₁ , σ₁ ⟩ ≜ ⟨ const● (C [ Thunk n₁ σ₁ ]) , n₂ , σ₂ ⟩

  ≜●    :  forall {n σ} -> ⟨ ● , n , σ ⟩ ≜ ⟨ ● , n , σ ⟩
  ≜●T   :  forall {C n₁ n₂ σ}
        -> ⟨ ● , n₁ , C [ Thunk n₂ σ ] ⟩ ≜ ⟨ Thunk● n₁ C , n₂ , σ ⟩
  ≜T●   :  forall {C n₁ n₂ σ}
        -> ⟨ Thunk● n₁ C , n₂ , σ ⟩ ≜ ⟨ ● , n₁ , C [ Thunk n₂ σ ] ⟩
  ≜TT   :  forall {C₁ C₂ n n₁ n₂ σ₁ σ₂}
        -> ⟨ C₁ , n₁ , σ₁ ⟩ ≜ ⟨ C₂ , n₂ , σ₂ ⟩
        -> ⟨ Thunk● n C₁ , n₁ , σ₁ ⟩ ≜ ⟨ Thunk● n C₂ , n₂ , σ₂ ⟩

  ≜●×²  :  forall {C₁ C₂ n₁ n₂ σ σ₁ σ₂}
        -> ⟨ C₁ , n₁ , σ₁ ⟩ ≜ ⟨ C₂ , n₂ , σ₂ ⟩
        -> ⟨ C₁ ●× σ , n₁ , σ₁ ⟩ ≜ ⟨ C₂ ●× σ , n₂ , σ₂ ⟩
  ≜×●²  :  forall {C₁ C₂ n₁ n₂ σ σ₁ σ₂}
        -> ⟨ C₁ , n₁ , σ₁ ⟩ ≜ ⟨ C₂ , n₂ , σ₂ ⟩
        -> ⟨ σ ×● C₁ , n₁ , σ₁ ⟩ ≜ ⟨ σ ×● C₂ , n₂ , σ₂ ⟩
  ≜●××● :  forall {C₁ C₂ n₁ n₂ σ₁ σ₂}
        -> ⟨ C₁ ●× C₂ [ Thunk n₂ σ₂ ] , n₁ , σ₁ ⟩ ≜
           ⟨ C₁ [ Thunk n₁ σ₁ ] ×● C₂ , n₂ , σ₂ ⟩
  ≜×●●× :  forall {C₁ C₂ n₁ n₂ σ₁ σ₂}
        -> ⟨ C₁ [ Thunk n₁ σ₁ ] ×● C₂ , n₂ , σ₂ ⟩ ≜
           ⟨ C₁ ●× C₂ [ Thunk n₂ σ₂ ] , n₁ , σ₁ ⟩

------------------------------------------------------------------------
-- The equality is meaningful

-- First some boring lemmas.

private

  drop-Thunkˡ :  forall {n₁ n₂ σ₁ σ₂}
             -> Thunk n₁ σ₁ ≡ Thunk n₂ σ₂ -> n₁ ≡ n₂
  drop-Thunkˡ ≡-refl = ≡-refl

  drop-Thunkʳ :  forall {n₁ n₂ σ₁ σ₂}
             -> Thunk n₁ σ₁ ≡ Thunk n₂ σ₂ -> σ₁ ≡ σ₂
  drop-Thunkʳ ≡-refl = ≡-refl

  drop-×ˡ :  forall {k} {σ₁¹ σ₂¹ σ₁² σ₂² : Ty k}
          -> σ₁¹ × σ₂¹ ≡ σ₁² × σ₂² -> σ₁¹ ≡ σ₁²
  drop-×ˡ ≡-refl = ≡-refl

  drop-×ʳ :  forall {k} {σ₁¹ σ₂¹ σ₁² σ₂² : Ty k}
          -> σ₁¹ × σ₂¹ ≡ σ₁² × σ₂² -> σ₂¹ ≡ σ₂²
  drop-×ʳ ≡-refl = ≡-refl

-- Soundness.

sound : forall {tt₁ tt₂} -> tt₁ ≜ tt₂ -> ⟦ tt₁ ⟧ ≡ ⟦ tt₂ ⟧
sound ≜c-               = ≡-refl
sound ≜-c               = ≡-refl
sound ≜●                = ≡-refl
sound ≜●T               = ≡-refl
sound ≜T●               = ≡-refl
sound (≜TT eq)          = ≡-cong (Thunk _) (sound eq)
sound (≜●×² {σ = σ} eq) = ≡-cong (\τ -> τ × σ) (sound eq)
sound (≜×●² {σ = σ} eq) = ≡-cong (\τ -> σ × τ) (sound eq)
sound ≜●××●             = ≡-refl
sound ≜×●●×             = ≡-refl

-- Completeness.

complete : forall tt₁ tt₂ -> ⟦ tt₁ ⟧ ≡ ⟦ tt₂ ⟧ -> tt₁ ≜ tt₂
complete ⟨ C₁ , _ , _ ⟩ ⟨ C₂ , _ , _ ⟩ eq = complete' C₁ C₂ eq
  where
  complete' :  forall C₁ {n₁ σ₁} C₂ {n₂ σ₂}
            -> C₁ [ Thunk n₁ σ₁ ] ≡ C₂ [ Thunk n₂ σ₂ ]
            -> ⟨ C₁ , n₁ , σ₁ ⟩ ≜ ⟨ C₂ , n₂ , σ₂ ⟩
  complete' (const● ._)    _              ≡-refl = ≜c-
  complete' _              (const● ._)    ≡-refl = ≜-c

  complete' ●              ●              ≡-refl = ≜●
  complete' ●              (Thunk● _ _)   ≡-refl = ≜●T
  complete' (Thunk● _ _)   ●              ≡-refl = ≜T●
  complete' (Thunk● n₁ C₁) (Thunk● n₂ C₂) eq with n₁ | drop-Thunkˡ eq
  ... | .n₂ | ≡-refl = ≜TT (complete' C₁ C₂ (drop-Thunkʳ eq))

  complete' (C₁ ●× σ₁) (C₂ ●× σ₂) eq with σ₁ | drop-×ʳ eq
  ... | .σ₂ | ≡-refl = ≜●×² (complete' C₁ C₂ (drop-×ˡ eq))
  complete' (σ₁ ×● C₁) (σ₂ ×● C₂) eq with σ₁ | drop-×ˡ eq
  ... | .σ₂ | ≡-refl = ≜×●² (complete' C₁ C₂ (drop-×ʳ eq))
  complete' (C₁ ●× ._) (._ ×● C₂) ≡-refl = ≜●××●
  complete' (._ ×● C₂) (C₁ ●× ._) ≡-refl = ≜×●●×

  complete' ●            (_ ●× _)     ()
  complete' ●            (_ ×● _)     ()
  complete' (Thunk● _ _) (_ ●× _)     ()
  complete' (Thunk● _ _) (_ ×● _)     ()
  complete' (_ ●× _)     ●            ()
  complete' (_ ●× _)     (Thunk● _ _) ()
  complete' (_ ×● _)     ●            ()
  complete' (_ ×● _)     (Thunk● _ _) ()
